package net.afyre.haha.demo.coq;

import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;

import net.afyre.haha.demo.solver.context.DeclarationVisitor;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;


/**
 * Declaration visitor used to convert declarations to text.
 *
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @author Aleksy Schubert <alx@mimuw.edu.pl>
 */
public final class CoqDeclarationPrinter implements DeclarationVisitor {


  /** Output stream. */
  private final Writer output;

  /** Object used to print expressions. */
  private final CoqExprPrinter exprPrinter;

  /** Controller of indentation in Coq. */
  private CoqIndent indent;

  /**
   * Create a visitor which will write all
   * inspected objects to an output stream.
   * 
   * @param output Output stream.
   * @param indent the indentation management object
   */
  public CoqDeclarationPrinter(final Writer output,
                               final CoqIndent indent) {
    this.output = output;
    this.exprPrinter = new CoqExprPrinter(output);
    this.indent = indent;
  }

  // TODO: Move this to utils
  /**
   * Print an object (using toString()) to the output stream,
   * converting any IOException to RuntimeExceptions.
   * @param obj Object to be printed (can be null).
   */
  private void print(final Object obj) {
    final String message;

    if (obj == null) {
      message = "null";
    } else {
      message = obj.toString();
    }
    try {
      output.write(message);
    } catch (final IOException ex) {
      throw new RuntimeException(ex);
    }
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitAxiom(final String name, final Formula axiom) {
    final String cname = CoqLanguage.coqifyName(name) + "A";
    print("(** Definition of the axiom " + name);
    if (!name.equals(cname)) {
      print("\n");
      indent.print();
      print("    - the Coq name is slightly different than in HAHA");
    }
    print(". *)\n");
    print("Axiom ");
    print(cname);
    print(" : ");
    axiom.accept(exprPrinter);
    print(".\n\n");
    print("(* The axiom is added to automated proof search procedures.\n");
    print("   Remove it to get higher credibility. *)\n");
    print("Hint Resolve " + cname + " : haha_hints.\n");
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitTermDeclaration(final String name,
                                final ValueType type) {
    print("Parameter ");
    print(CoqLanguage.coqifyName(name));
    print(" : ");
    type.accept(exprPrinter);
    print(".\n");
  }

  /** {@inheritDoc}
   */
  @Override
  public void visitTermDefinition(final String name,
                                  final ValueType type,
                                  final Term value) {
    print("Parameter ");
    print(name);
    print(" : ");
    type.accept(exprPrinter);
    print(" = ");
    value.accept(exprPrinter);
    print(".");
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitPropDeclaration(final String name) {
    print("Parameter ");
    print(CoqLanguage.coqifyName(name));
    print(" : ");
    print("Prop.");
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitPropDefinition(final String name,
                                  final Formula value) {
    print("CONST ");
    print(CoqLanguage.coqifyName(name));
    print(" = ");
    value.accept(exprPrinter);
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitFunctionDeclaration(final String name,
                                       final Iterable<VariableBinding> args,
                                       final ValueType result) {
    final int pos = name.indexOf("function@");
    String nname = name;
    if (pos >= 0) {
      nname = name.substring(pos + "function@".length());
    }
    final String cname = CoqLanguage.coqifyName(nname) + "F";
    print("(** Declaration of the function name " + nname);
    if (!nname.equals(cname)) {
      print("\n");
      indent.print();
      print("    - the Coq name is slightly different than in HAHA");
    }
    print(". *)\n");
    indent.print();
    print("Parameter ");
    print(cname);
    print(" : ");

    /* Print arguments */
    final Iterator<VariableBinding> iter = args.iterator();
    while (iter.hasNext()) {
      final VariableBinding arg = iter.next();
      arg.getType().accept(exprPrinter);
      print(" -> ");
    }
    result.accept(exprPrinter);
    print(".\n");
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitFunctionDefinition(final String name,
                                      final Iterable<VariableBinding> args,
                                      final ValueType result, final Term body) {
    final String cname = CoqLanguage.coqifyName(name);

    if (body == null) {
      visitFunctionDeclaration(name, args, result);
      return;
    } else {
      print("(** Definition of the function " + name);
    }
    if (!name.equals(cname)) {
      print("\n");
      indent.print();
      print("    - the Coq name is slightly different than in HAHA");
    }
    print(". *)\n");
    indent.print();

    print("Definition " + cname + " : ");
    final Iterator<VariableBinding> iter = args.iterator();
    while (iter.hasNext()) {
      final VariableBinding arg = iter.next();
      print(arg.getName());
      print(" : ");
      arg.getType().accept(exprPrinter);
      if (iter.hasNext()) {
        print(", ");
      }
    }
    print(" -> ");
    result.accept(exprPrinter);
    print(" = ");
    body.accept(exprPrinter);
    print(".");
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitPredicateDeclaration(
      final String name,
      final Iterable<VariableBinding> args) {
    visitPredicateDefinition(name, args, null);
  }

  /** {@inheritDoc} 
   */
  @Override
  public void visitPredicateDefinition(final String name,
                                       final Iterable<VariableBinding> args,
                                       final Formula body) {
    final String cname = CoqLanguage.coqifyName(name);
    print("(** Definition of the predicate " + name);
    if (!name.equals(cname)) {
      print("\n");
      indent.print();
      print("    - the Coq name is slightly different than in HAHA");
    }
    print(". *)\n");
    indent.print();
    print("Definition ");
    print(cname);

    /* Print arguments */
    final Iterator<VariableBinding> iter = args.iterator();
    while (iter.hasNext()) {
      final VariableBinding arg = iter.next();
      print("(");
      print(CoqLanguage.coqifyName(arg.getName()));
      print(" : ");
      arg.getType().accept(exprPrinter);
      print(")");
      if (iter.hasNext()) {
        print(" ");
      }
    }
    print(" : Prop ");

    /* Print body of the definition */
    if (body != null) {
      print(":=\n");
      indent.inc();
      indent.print();
      body.accept(exprPrinter);
      indent.dec();
    }
    print(".\n\n");
    print("(* We allow this predicate to be automatically unfolded by our " +
          "automatic\n" +
          "   tactics. *)\n");
    print("Hint Unfold " + cname + ": haha_predicates.\n");
  }

}
