package net.afyre.haha.demo.solver.smt;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;
import net.afyre.haha.demo.solver.smt.sexpr.DefaultSExprFactory;
import net.afyre.haha.demo.solver.smt.sexpr.SExpr;
import net.afyre.haha.demo.solver.smt.sexpr.SExprFactory;

/**
 * Visitor used to convert terms and, types and formulae to
 * S-Expressions in SMTv2 format.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class SExprBuilder implements ExpressionSwitch<SExpr> {
  /** Factory used to build S-Expressions. */
  private static final SExprFactory factory = new DefaultSExprFactory();

  /**
   * Create a list.
   * @param elements Elements.
   * @return List as S-expression.
   */
  private SExpr list(final SExpr... elements) {
    return factory.createList(Arrays.asList(elements));
  }
  
  /**
   * Create a list.
   * @param op first element (will be converted to a symbol).
   * @param arguments Other elements.
   * @return List as S-expression.
   */
  private SExpr apply(final String op, 
                      final SExpr... arguments) {
    final List<SExpr> elements = new ArrayList<SExpr>();
    elements.add(factory.createSymbol(op));
    for (final SExpr argument : arguments) {
      elements.add(argument);
    }
    return factory.createList(elements);
  }
  
  
  /** {@inheritDoc} */
  @Override
  public SExpr caseIntConst(final BigInteger value) {
    return factory.createInteger(value);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVariable(final String var) {
    return factory.createSymbol(var);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseMkTuple(final Term... elements) {
    throw new UnsupportedOperationException("Tuples are not supported.");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseTupleAccess(final Term tuple, final int index) {
    throw new UnsupportedOperationException("Tuples are not supported.");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseUminus(final Term arg) {
    return apply("-", arg.accept(this));
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseAddition(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("+", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseSubtraction(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("-", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseMultiplication(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("*", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseDivision(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("div", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseRemainder(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("mod", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr casePower(final Term left, final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("^", leftExpr, rightExpr);
  }
  
  /** {@inheritDoc} */
  @Override
  public SExpr caseLet(final String var, 
                       final ValueType type, 
                       final Term value, 
                       final Term body) {
    final SExpr varExpr = factory.createSymbol(var);
    
    final SExpr valueExpr = value.accept(this);
    final SExpr bodyExpr = body.accept(this);
    
    final SExpr bindings = list(list(varExpr, valueExpr));
    
    return apply("let", bindings, bodyExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseLetProp(final String var, 
                           final Formula value, 
                           final Term body) {
    final SExpr varExpr = factory.createSymbol(var);
    final SExpr valueExpr = value.accept(this);
    final SExpr bodyExpr = body.accept(this);
   
    final SExpr bindings = list(list(varExpr, valueExpr));
    
    return apply("let", bindings, bodyExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseIfThenElse(final Formula cond, 
                              final Term ifTrue, 
                              final Term ifFalse) {
    
    final SExpr condExpr = cond.accept(this);
    final SExpr trueExpr = ifTrue.accept(this);
    
    final SExpr falseExpr = ifFalse.accept(this);
    return apply("ite", condExpr, trueExpr, falseExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseArrayAccess(final Term arr, 
                               final Term index) {
    
    final SExpr arrExpr = arr.accept(this);
    final SExpr indexExpr = index.accept(this);
    return apply("select", arrExpr, indexExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseArrayUpdate(final Term arr, 
                               final Term index, 
                               final Term value) {
    
    final SExpr arrExpr = arr.accept(this);
    final SExpr indexExpr = index.accept(this);
    final SExpr valueExpr = value.accept(this);
    return apply("store", arrExpr, indexExpr, valueExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseCall(final String func, 
                        final Term... args) {
    final SExpr[] argExprs = new SExpr[args.length];
    for (int i = 0; i < args.length; i++) {
      argExprs[i] = args[i].accept(this);
    }
    return apply(func, argExprs);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseConcat(final Term left, 
                          final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("concat", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseSelectBits(final Term vec, 
                              final int left, 
                              final int right) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr leftExpr = 
        factory.createInteger(BigInteger.valueOf(left));
    final SExpr rightExpr = 
        factory.createInteger(BigInteger.valueOf(right));
    final SExpr extractExpr = apply("extract", leftExpr, rightExpr);
    return list(extractExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseShl(final Term vec, 
                       final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createBin(BigInteger.valueOf(offset), 0);
    return apply("bvshl", vecExpr, offsetExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseShr(final Term vec, 
                       final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createBin(BigInteger.valueOf(offset), 0);
    return apply("bvlshr", vecExpr, offsetExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseSX(final Term vec, 
                      final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createInteger(BigInteger.valueOf(offset));
    final SExpr opExpr = apply("sign_extend", offsetExpr);
    return list(opExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseZeroExtend(final Term vec, 
                              final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createInteger(BigInteger.valueOf(offset));
    final SExpr opExpr = apply("zero_extend", offsetExpr);
    return list(opExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseRepeat(final Term vec, 
                          final int count) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr countExpr = 
        factory.createInteger(BigInteger.valueOf(count));
    final SExpr opExpr = apply("repeat", countExpr);
    return list(opExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseRotateLeft(final Term vec, 
                              final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createInteger(BigInteger.valueOf(offset));
    final SExpr opExpr = apply("rotate_left", offsetExpr);
    return list(opExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseRotateRight(final Term vec, 
                               final int offset) {
    
    final SExpr vecExpr = vec.accept(this);
    final SExpr offsetExpr = 
        factory.createInteger(BigInteger.valueOf(offset));
    final SExpr opExpr = apply("rotate_right", offsetExpr);
    return list(opExpr, vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorAnd(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("bvand", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorOr(final Term left, 
                            final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("bvor", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorXor(final Term left, 
                             final Term right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("bvxor", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorNand(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvnand", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorNor(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvnor", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorXnor(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvxnor", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorCmp(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvcomp", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSum(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvadd", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorDiff(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsub", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorUminus(final Term vec) {
    
    final SExpr vecExpr = vec.accept(this);
    return apply("bvneg", vecExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorProd(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvmul", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorShl(final Term vec, 
                             final Term offset) {
    
    final SExpr vecExpr = vec.accept(this);
    
    final SExpr offsetExpr = offset.accept(this);
    return apply("bvshl", vecExpr, offsetExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorLshr(final Term vec, 
                              final Term offset) {
    
    final SExpr vecExpr = vec.accept(this);
    
    final SExpr offsetExpr = offset.accept(this);
    return apply("bvlshr", vecExpr, offsetExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorRshr(final Term vec, 
                              final Term offset) {
    
    final SExpr vecExpr = vec.accept(this);
    
    final SExpr offsetExpr = offset.accept(this);
    return apply("bvashr", vecExpr, offsetExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorUdiv(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvudiv", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSdiv(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsdiv", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorUrem(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvurem", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSrem(final Term left, 
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsrem", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSmod(final Term left,
                              final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsmod", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorLt(final Term left, 
                            final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvult", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorGt(final Term left, 
                            final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvugt", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorLe(final Term left, 
                            final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvule", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorGe(final Term left, 
                            final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvuge", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSlt(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvslt", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSgt(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsgt", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSle(final Term left, 
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsle", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseVectorSge(final Term left,
                             final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("bvsge", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseTrue() {
    return factory.createSymbol("true");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseFalse() {
    return factory.createSymbol("false");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseNegation(final Formula arg) {
    return apply("not", arg.accept(this));
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseConjunction(final Formula left, 
                               final Formula right) {
    final SExpr leftExpr = left.accept(this);
    final SExpr rightExpr = right.accept(this);
    return apply("and", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseDisjunction(final Formula left, 
                               final Formula right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("or", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseImplication(final Formula left, 
                               final Formula right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("=>", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseXor(final Formula left, final Formula right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("xor", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseEquivalence(final Formula left, 
                               final Formula right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("=", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseEq(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("=", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseNeq(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("distinct", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseDistinct(final Term... args) {
    final SExpr[] argExprs = new SExpr[args.length];
    
    for (int i = 0; i < args.length; i++) {
      argExprs[i] = args[i].accept(this);
    }
    return apply("distinct", argExprs);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseLt(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("<", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseGt(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply(">", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseLe(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply("<=", leftExpr, rightExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseGe(final Term left, final Term right) {
    
    final SExpr leftExpr = left.accept(this);
    
    final SExpr rightExpr = right.accept(this);
    return apply(">=", leftExpr, rightExpr);
  }

  /** Process a quantified formula. 
   * @param quantifier Quantifier name.
   * @param variables Bound variables.
   * @param arg Formula body.
   * @param patterns Instantiation patterns. 
   * @return Transformed expression.
   **/
  public SExpr caseQuantifiedFormula(
      final String quantifier,
      final Iterable<VariableBinding> variables,
      final Formula arg,
      final Expression... patterns) {
    final List<SExpr> varExprs = new ArrayList<SExpr>();
    for (final VariableBinding variable : variables) {
      final SExpr typeExpr = variable.getType().accept(this);
      final SExpr bindingExpr = apply(variable.getName(), typeExpr);
      varExprs.add(bindingExpr);
    }
    final SExpr bindingsExpr = factory.createList(varExprs);   
    SExpr argExpr = arg.accept(this);
    if (patterns.length > 0) {
      final List<SExpr> patternExprs = new ArrayList<SExpr>();  
      for (final Expression pattern : patterns) {
        patternExprs.add(pattern.accept(this));
      }
      argExpr = apply("!", argExpr, 
                      factory.createKeyword("pattern"), 
                      factory.createList(patternExprs));
    }
    return apply(quantifier, bindingsExpr, argExpr);
  }
  
  /** {@inheritDoc} */
  @Override
  public SExpr caseExists(final Iterable<VariableBinding> variables,
                          final Formula arg,
                          final Expression... patterns) {
    return caseQuantifiedFormula("exists", variables, arg, patterns);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseForall(final Iterable<VariableBinding> variables,
                          final Formula arg, 
                          final Expression... patterns) {

    return caseQuantifiedFormula("forall", variables, arg, patterns);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseFLet(final String var, 
                        final ValueType type, 
                        final Term value, 
                        final Formula body) {
    final SExpr varExpr = factory.createSymbol(var);
    final SExpr valueExpr = value.accept(this);    
    final SExpr bodyExpr = body.accept(this);

    final SExpr bindings = list(list(varExpr, valueExpr));
    
    return apply("let", bindings, bodyExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseFLetProp(final String var, 
                            final Formula value, 
                            final Formula body) {
    final SExpr varExpr = factory.createSymbol(var);
    final SExpr valueExpr = value.accept(this);    
    final SExpr bodyExpr = body.accept(this);

    final SExpr bindings = list(list(varExpr, valueExpr));
    
    return apply("let", bindings, bodyExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseRational() {
    return factory.createSymbol("Real");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseInt() {
    return factory.createSymbol("Int");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseBitvector(final int size) {
    return apply("_", 
          factory.createSymbol("bv"), 
          factory.createInteger(BigInteger.valueOf(size)));
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseArray(final ValueType indexType,
                         final ValueType valueType) {
    
    final SExpr indexExpr = indexType.accept(this);
    
    final SExpr valueExpr = valueType.accept(this);
    
    return apply("Array", indexExpr, valueExpr);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseTuple(final ValueType... types) {
    throw new UnsupportedOperationException("Tuples are not supported.");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseFormulaVariable(final String name) {
    return factory.createSymbol(name);
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseProp() {
    return factory.createSymbol("Bool");
  }

  /** {@inheritDoc} */
  @Override
  public SExpr caseFunction(final ValueType resultType, 
                            final ValueType... argTypes) {
    throw new UnsupportedOperationException();
  }

  /** {@inheritDoc} */
  @Override
  public SExpr casePredicate(final ValueType... argTypes) {
    throw new UnsupportedOperationException();
  }

  /** {@inheritDoc} */
  @Override
  public SExpr casePredicateCall(final String pred, 
                                 final Term... args) {
    final List<SExpr> result = new ArrayList<SExpr>();
    result.add(factory.createSymbol(pred));
    for (final Term arg : args) {
      result.add(arg.accept(this));
    }
    return factory.createList(result);
  }
}
