package net.afyre.haha.demo.solver.dummy;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import net.afyre.haha.demo.solver.context.Answer;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.context.DeclarationPrinter;
import net.afyre.haha.demo.solver.context.DeclarationVisitor;
import net.afyre.haha.demo.solver.context.SolverContext;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.util.ExprPrinter;

/**
 * Solver implementation which prints 
 * all received commands to a specified stream
 * and passes them to another solver.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public class VerboseSolver implements SolverContext {
  /** Wrapped solver. */
  private final SolverContext solver;
  
  /** Output stream. */
  private final Writer output;
  
  /** Visitor used to print expressions. */
  private final ExpressionSwitch<Void> exprPrinter;
  
  /** Visitor used to print declarations. */
  private final DeclarationVisitor declPrinter;
  
  /** Declaration stack. Marks are represented by null values. */
  private final Stack<Declaration> declarations;
  
  /**
   * Create a solver wrapper which prints all commands 
   * to an output stream.
   * @param solver Solver which will be used to execute commands.
   * @param output Output stream.
   */
  public VerboseSolver(final SolverContext solver,
                       final Writer output) {
    this.solver = solver;
    this.output = output;
    this.exprPrinter = new ExprPrinter(output);
    this.declPrinter = new DeclarationPrinter(output);
    this.declarations = new Stack<Declaration>();
  }

  /**
   * Create a solver wrapper which prints all commands 
   * to stderr.
   * @param solver Solver which will be used to execute commands.
   */
  public VerboseSolver(final SolverContext solver) {
    this.solver = solver;
    this.output = new OutputStreamWriter(System.err);
    this.exprPrinter = new ExprPrinter(output);
    this.declPrinter = new DeclarationPrinter(output);
    this.declarations = new Stack<Declaration>();
  }
  
  /**
   * Create a dummy solver which prints all commands 
   * to stderr and always returns UNKNOWN.
   */
  public VerboseSolver() {
    this.solver = new DummySolver();
    this.output = new OutputStreamWriter(System.err);
    this.exprPrinter = new ExprPrinter(output);
    this.declPrinter = new DeclarationPrinter(output);
    this.declarations = new Stack<Declaration>();
  }
  
  /**
   * Print a string (using toString()) to the output stream, 
   * converting any IOException to RuntimeExceptions.
   * @param msg String to be printed (can be null).
   */
  private void print(final String msg) {
    final String message;
    
    if (msg == null) {
      message = "null";
    } else {
      message = msg;
    }
    try {
      output.write(message);
    } catch (final IOException e) {
      throw new RuntimeException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public boolean addDeclaration(final Declaration declaration) {
    declarations.push(declaration);
    print("DECLARE ");
    declaration.accept(declPrinter);
    print("\n");
    
    return solver.addDeclaration(declaration);
  }
  
  /** {@inheritDoc} */
  @Override
  public boolean mark() {
    declarations.push(null);
    print("MARK\n");
    
    return solver.mark();
  }
  
  /** {@inheritDoc} */
  @Override  
  public boolean reset() {
    // Collect list of removed symbols.
    final List<String> removedSymbols = new ArrayList<String>();
    
    boolean markFound = false;
    while (!declarations.empty() && !markFound) {
      final Declaration decl = declarations.pop();
      if (decl == null) {
        markFound = true;
      } else {
        removedSymbols.add(decl.getName());
      }
    }
    
    print("RESET - ");
    
    if (removedSymbols.size() > 0) {
      print("Removed symbols:");
      for (final String symbol : removedSymbols) {
        print(" ");
        print(symbol);
      }
    } else {
      print("No symbols removed.");
    }
    print("\n");
    return solver.reset();
  }
  
  /** {@inheritDoc} */
  @Override
  public Answer isValid(final Formula query) {
    print("IS_VALID ");
    query.accept(exprPrinter);
    print("\n");
    
    final Answer answer = solver.isValid(query);
    
    print("Result: " + answer.toString() + "\n");
    
    return answer;
  }

  /** {@inheritDoc} */
  @Override
  public Answer isSatisfiable(final Formula query) {
    print("IS_VALID ");
    query.accept(exprPrinter);
    print("\n");
    final Answer answer = solver.isSatisfiable(query);
    
    print("Result: " + answer.toString() + "\n");
    
    return answer;
  }

  /** {@inheritDoc} */
  @Override
  public void finish() {
    print("Releasing solver resources... ");
    solver.finish();
    print("done\n");
  }
}
