package net.afyre.haha.demo.solver.expr.terms;

import java.math.BigInteger;

import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Interface of objects used to create terms.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public interface TermFactory {
  /**
   * Create a term of kind 'IntConst'
   * @see TermSwitch#caseIntConst
   * @param value {@link TermSwitch#caseIntConst}
   * @return Expression object.
   */
  Term createIntConst(BigInteger value);

  /**
   * Create a term of kind 'Variable'
   * @see TermSwitch#caseVariable
   * @param var {@link TermSwitch#caseVariable}
   * @return Expression object.
   */
  Term createVariable(String var);

  /**
   * Create a term of kind 'MkTuple'
   * @see TermSwitch#caseMkTuple
   * @param elements {@link TermSwitch#caseMkTuple}
   * @return Expression object.
   */
  Term createMkTuple(Term... elements);

  /**
   * Create a term of kind 'TupleAccess'
   * @see TermSwitch#caseTupleAccess
   * @param tuple {@link TermSwitch#caseTupleAccess}
   * @param index {@link TermSwitch#caseTupleAccess}
   * @return Expression object.
   */
  Term createTupleAccess(Term tuple,
                         int index);

  /**
   * Create a term of kind 'Uminus'
   * @see TermSwitch#caseUminus
   * @param arg {@link TermSwitch#caseUminus}
   * @return Expression object.
   */
  Term createUminus(Term arg);

  /**
   * Create a term of kind 'Addition'
   * @see TermSwitch#caseAddition
   * @param left {@link TermSwitch#caseAddition}
   * @param right {@link TermSwitch#caseAddition}
   * @return Expression object.
   */
  Term createAddition(Term left,
                      Term right);

  /**
   * Create a term of kind 'Subtraction'
   * @see TermSwitch#caseSubtraction
   * @param left {@link TermSwitch#caseSubtraction}
   * @param right {@link TermSwitch#caseSubtraction}
   * @return Expression object.
   */
  Term createSubtraction(Term left,
                         Term right);

  /**
   * Create a term of kind 'Multiplication'
   * @see TermSwitch#caseMultiplication
   * @param left {@link TermSwitch#caseMultiplication}
   * @param right {@link TermSwitch#caseMultiplication}
   * @return Expression object.
   */
  Term createMultiplication(Term left,
                            Term right);

  /**
   * Create a term of kind 'Division'
   * @see TermSwitch#caseDivision
   * @param left {@link TermSwitch#caseDivision}
   * @param right {@link TermSwitch#caseDivision}
   * @return Expression object.
   */
  Term createDivision(Term left,
                      Term right);

  /**
   * Create a term of kind 'Remainder'
   * @see TermSwitch#caseRemainder
   * @param left {@link TermSwitch#caseRemainder}
   * @param right {@link TermSwitch#caseRemainder}
   * @return Expression object.
   */
  Term createRemainder(Term left,
                       Term right);

  /**
   * Create a term of kind 'Power'
   * @see TermSwitch#casePower
   * @param left {@link TermSwitch#casePower}
   * @param right {@link TermSwitch#casePower}
   * @return Expression object.
   */
  Term createPower(Term left,
                   Term right);

  /**
   * Create a term of kind 'Let'
   * @see TermSwitch#caseLet
   * @param var {@link TermSwitch#caseLet}
   * @param type {@link TermSwitch#caseLet}
   * @param value {@link TermSwitch#caseLet}
   * @param body {@link TermSwitch#caseLet}
   * @return Expression object.
   */
  Term createLet(String var,
                 ValueType type,
                 Term value,
                 Term body);

  /**
   * Create a term of kind 'LetProp'
   * @see TermSwitch#caseLetProp
   * @param var {@link TermSwitch#caseLetProp}
   * @param value {@link TermSwitch#caseLetProp}
   * @param body {@link TermSwitch#caseLetProp}
   * @return Expression object.
   */
  Term createLetProp(String var,
                     Formula value,
                     Term body);

  /**
   * Create a term of kind 'IfThenElse'
   * @see TermSwitch#caseIfThenElse
   * @param cond {@link TermSwitch#caseIfThenElse}
   * @param ifTrue {@link TermSwitch#caseIfThenElse}
   * @param ifFalse {@link TermSwitch#caseIfThenElse}
   * @return Expression object.
   */
  Term createIfThenElse(Formula cond,
                        Term ifTrue,
                        Term ifFalse);

  /**
   * Create a term of kind 'ArrayAccess'
   * @see TermSwitch#caseArrayAccess
   * @param arr {@link TermSwitch#caseArrayAccess}
   * @param index {@link TermSwitch#caseArrayAccess}
   * @return Expression object.
   */
  Term createArrayAccess(Term arr,
                         Term index);

  /**
   * Create a term of kind 'ArrayUpdate'
   * @see TermSwitch#caseArrayUpdate
   * @param arr {@link TermSwitch#caseArrayUpdate}
   * @param index {@link TermSwitch#caseArrayUpdate}
   * @param value {@link TermSwitch#caseArrayUpdate}
   * @return Expression object.
   */
  Term createArrayUpdate(Term arr,
                         Term index,
                         Term value);

  /**
   * Create a term of kind 'Call'
   * @see TermSwitch#caseCall
   * @param func {@link TermSwitch#caseCall}
   * @param args {@link TermSwitch#caseCall}
   * @return Expression object.
   */
  Term createCall(String func,
                  Term... args);

  /**
   * Create a term of kind 'Concat'
   * @see TermSwitch#caseConcat
   * @param left {@link TermSwitch#caseConcat}
   * @param right {@link TermSwitch#caseConcat}
   * @return Expression object.
   */
  Term createConcat(Term left,
                    Term right);

  /**
   * Create a term of kind 'SelectBits'
   * @see TermSwitch#caseSelectBits
   * @param vec {@link TermSwitch#caseSelectBits}
   * @param left {@link TermSwitch#caseSelectBits}
   * @param right {@link TermSwitch#caseSelectBits}
   * @return Expression object.
   */
  Term createSelectBits(Term vec,
                        int left,
                        int right);

  /**
   * Create a term of kind 'Shl'
   * @see TermSwitch#caseShl
   * @param vec {@link TermSwitch#caseShl}
   * @param offset {@link TermSwitch#caseShl}
   * @return Expression object.
   */
  Term createShl(Term vec,
                 int offset);

  /**
   * Create a term of kind 'Shr'
   * @see TermSwitch#caseShr
   * @param vec {@link TermSwitch#caseShr}
   * @param offset {@link TermSwitch#caseShr}
   * @return Expression object.
   */
  Term createShr(Term vec,
                 int offset);

  /**
   * Create a term of kind 'SX'
   * @see TermSwitch#caseSX
   * @param vec {@link TermSwitch#caseSX}
   * @param offset {@link TermSwitch#caseSX}
   * @return Expression object.
   */
  Term createSX(Term vec,
                int offset);

  /**
   * Create a term of kind 'ZeroExtend'
   * @see TermSwitch#caseZeroExtend
   * @param vec {@link TermSwitch#caseZeroExtend}
   * @param offset {@link TermSwitch#caseZeroExtend}
   * @return Expression object.
   */
  Term createZeroExtend(Term vec,
                        int offset);

  /**
   * Create a term of kind 'Repeat'
   * @see TermSwitch#caseRepeat
   * @param vec {@link TermSwitch#caseRepeat}
   * @param count {@link TermSwitch#caseRepeat}
   * @return Expression object.
   */
  Term createRepeat(Term vec,
                    int count);

  /**
   * Create a term of kind 'RotateLeft'
   * @see TermSwitch#caseRotateLeft
   * @param vec {@link TermSwitch#caseRotateLeft}
   * @param offset {@link TermSwitch#caseRotateLeft}
   * @return Expression object.
   */
  Term createRotateLeft(Term vec,
                        int offset);

  /**
   * Create a term of kind 'RotateRight'
   * @see TermSwitch#caseRotateRight
   * @param vec {@link TermSwitch#caseRotateRight}
   * @param offset {@link TermSwitch#caseRotateRight}
   * @return Expression object.
   */
  Term createRotateRight(Term vec,
                         int offset);

  /**
   * Create a term of kind 'VectorAnd'
   * @see TermSwitch#caseVectorAnd
   * @param left {@link TermSwitch#caseVectorAnd}
   * @param right {@link TermSwitch#caseVectorAnd}
   * @return Expression object.
   */
  Term createVectorAnd(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorOr'
   * @see TermSwitch#caseVectorOr
   * @param left {@link TermSwitch#caseVectorOr}
   * @param right {@link TermSwitch#caseVectorOr}
   * @return Expression object.
   */
  Term createVectorOr(Term left,
                      Term right);

  /**
   * Create a term of kind 'VectorXor'
   * @see TermSwitch#caseVectorXor
   * @param left {@link TermSwitch#caseVectorXor}
   * @param right {@link TermSwitch#caseVectorXor}
   * @return Expression object.
   */
  Term createVectorXor(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorNand'
   * @see TermSwitch#caseVectorNand
   * @param left {@link TermSwitch#caseVectorNand}
   * @param right {@link TermSwitch#caseVectorNand}
   * @return Expression object.
   */
  Term createVectorNand(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorNor'
   * @see TermSwitch#caseVectorNor
   * @param left {@link TermSwitch#caseVectorNor}
   * @param right {@link TermSwitch#caseVectorNor}
   * @return Expression object.
   */
  Term createVectorNor(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorXnor'
   * @see TermSwitch#caseVectorXnor
   * @param left {@link TermSwitch#caseVectorXnor}
   * @param right {@link TermSwitch#caseVectorXnor}
   * @return Expression object.
   */
  Term createVectorXnor(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorCmp'
   * @see TermSwitch#caseVectorCmp
   * @param left {@link TermSwitch#caseVectorCmp}
   * @param right {@link TermSwitch#caseVectorCmp}
   * @return Expression object.
   */
  Term createVectorCmp(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorSum'
   * @see TermSwitch#caseVectorSum
   * @param left {@link TermSwitch#caseVectorSum}
   * @param right {@link TermSwitch#caseVectorSum}
   * @return Expression object.
   */
  Term createVectorSum(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorDiff'
   * @see TermSwitch#caseVectorDiff
   * @param left {@link TermSwitch#caseVectorDiff}
   * @param right {@link TermSwitch#caseVectorDiff}
   * @return Expression object.
   */
  Term createVectorDiff(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorUminus'
   * @see TermSwitch#caseVectorUminus
   * @param vec {@link TermSwitch#caseVectorUminus}
   * @return Expression object.
   */
  Term createVectorUminus(Term vec);

  /**
   * Create a term of kind 'VectorProd'
   * @see TermSwitch#caseVectorProd
   * @param left {@link TermSwitch#caseVectorProd}
   * @param right {@link TermSwitch#caseVectorProd}
   * @return Expression object.
   */
  Term createVectorProd(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorShl'
   * @see TermSwitch#caseVectorShl
   * @param vec {@link TermSwitch#caseVectorShl}
   * @param offset {@link TermSwitch#caseVectorShl}
   * @return Expression object.
   */
  Term createVectorShl(Term vec,
                       Term offset);

  /**
   * Create a term of kind 'VectorLshr'
   * @see TermSwitch#caseVectorLshr
   * @param vec {@link TermSwitch#caseVectorLshr}
   * @param offset {@link TermSwitch#caseVectorLshr}
   * @return Expression object.
   */
  Term createVectorLshr(Term vec,
                        Term offset);

  /**
   * Create a term of kind 'VectorRshr'
   * @see TermSwitch#caseVectorRshr
   * @param vec {@link TermSwitch#caseVectorRshr}
   * @param offset {@link TermSwitch#caseVectorRshr}
   * @return Expression object.
   */
  Term createVectorRshr(Term vec,
                        Term offset);

  /**
   * Create a term of kind 'VectorUdiv'
   * @see TermSwitch#caseVectorUdiv
   * @param left {@link TermSwitch#caseVectorUdiv}
   * @param right {@link TermSwitch#caseVectorUdiv}
   * @return Expression object.
   */
  Term createVectorUdiv(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorSdiv'
   * @see TermSwitch#caseVectorSdiv
   * @param left {@link TermSwitch#caseVectorSdiv}
   * @param right {@link TermSwitch#caseVectorSdiv}
   * @return Expression object.
   */
  Term createVectorSdiv(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorUrem'
   * @see TermSwitch#caseVectorUrem
   * @param left {@link TermSwitch#caseVectorUrem}
   * @param right {@link TermSwitch#caseVectorUrem}
   * @return Expression object.
   */
  Term createVectorUrem(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorSrem'
   * @see TermSwitch#caseVectorSrem
   * @param left {@link TermSwitch#caseVectorSrem}
   * @param right {@link TermSwitch#caseVectorSrem}
   * @return Expression object.
   */
  Term createVectorSrem(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorSmod'
   * @see TermSwitch#caseVectorSmod
   * @param left {@link TermSwitch#caseVectorSmod}
   * @param right {@link TermSwitch#caseVectorSmod}
   * @return Expression object.
   */
  Term createVectorSmod(Term left,
                        Term right);

  /**
   * Create a term of kind 'VectorLt'
   * @see TermSwitch#caseVectorLt
   * @param left {@link TermSwitch#caseVectorLt}
   * @param right {@link TermSwitch#caseVectorLt}
   * @return Expression object.
   */
  Term createVectorLt(Term left,
                      Term right);

  /**
   * Create a term of kind 'VectorGt'
   * @see TermSwitch#caseVectorGt
   * @param left {@link TermSwitch#caseVectorGt}
   * @param right {@link TermSwitch#caseVectorGt}
   * @return Expression object.
   */
  Term createVectorGt(Term left,
                      Term right);

  /**
   * Create a term of kind 'VectorLe'
   * @see TermSwitch#caseVectorLe
   * @param left {@link TermSwitch#caseVectorLe}
   * @param right {@link TermSwitch#caseVectorLe}
   * @return Expression object.
   */
  Term createVectorLe(Term left,
                      Term right);

  /**
   * Create a term of kind 'VectorGe'
   * @see TermSwitch#caseVectorGe
   * @param left {@link TermSwitch#caseVectorGe}
   * @param right {@link TermSwitch#caseVectorGe}
   * @return Expression object.
   */
  Term createVectorGe(Term left,
                      Term right);

  /**
   * Create a term of kind 'VectorSlt'
   * @see TermSwitch#caseVectorSlt
   * @param left {@link TermSwitch#caseVectorSlt}
   * @param right {@link TermSwitch#caseVectorSlt}
   * @return Expression object.
   */
  Term createVectorSlt(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorSgt'
   * @see TermSwitch#caseVectorSgt
   * @param left {@link TermSwitch#caseVectorSgt}
   * @param right {@link TermSwitch#caseVectorSgt}
   * @return Expression object.
   */
  Term createVectorSgt(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorSle'
   * @see TermSwitch#caseVectorSle
   * @param left {@link TermSwitch#caseVectorSle}
   * @param right {@link TermSwitch#caseVectorSle}
   * @return Expression object.
   */
  Term createVectorSle(Term left,
                       Term right);

  /**
   * Create a term of kind 'VectorSge'
   * @see TermSwitch#caseVectorSge
   * @param left {@link TermSwitch#caseVectorSge}
   * @param right {@link TermSwitch#caseVectorSge}
   * @return Expression object.
   */
  Term createVectorSge(Term left,
                       Term right);
}
