package net.afyre.haha.demo.solver.expr.util;

import java.util.HashMap;
import java.util.Map;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.impl.SimpleExpressionFactory;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * A mapping from variables to expressions.
 * Mappings for term and formula variables are kept separately.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class Substitution {
  /** Term variable mappings. */
  private final Map<String, Term> termMap;
  
  /** Formula variable mappings. */
  private final Map<String, Formula> formulaMap;
  
  /**
   * Create a new substitution, initialized as identity.
   */
  public Substitution() {
    termMap = new HashMap<String, Term>();
    formulaMap = new HashMap<String, Formula>();
  }
 
  /**
   * Change the value assigned to a term variable.
   * @param var Variable name.
   * @param value New value.
   */
  public void setTermBinding(final String var, final Term value) {
    termMap.put(var, value);
  }

  /**
   * Change the value assigned to a formula variable.
   * @param var Variable name.
   * @param value New value.
   */
  public void setFormulaBinding(final String var, final Formula value) {
    formulaMap.put(var, value);
  }
  
  /**
   * Apply the substitution to a given expression.
   * All substitutions are performed in parallel, i.e. 
   * variables contained in substituted values are
   * never expanded. 
   * Bound variables are not modified.
   * @param expr Source expression.
   * @return Expression after substitution.
   */
  public Expression perform(final Expression expr) {
    // Ughhh...
    if (expr instanceof Formula) {
      return perform((Formula)expr);
    } else if (expr instanceof Term) {
      return perform((Term)expr);
    } else {
      return expr;
    }
  }

  /**
   * Apply the substitution to a given term.
   * See {@link #perform(Expression)} for details.
   * @param term Source term.
   * @return Term after substitution.
   */
  public Term perform(final Term term) {
    return term.accept(new Worker());
  }
  
  /**
   * Apply the substitution to a given formula.
   * See {@link #perform(Expression)} for details.
   * @param formula Source formula.
   * @return Formula after substitution.
   */  
  public Formula perform(final Formula formula) {
    return formula.accept(new Worker()); 
  }
  
  /** 
   * Worker class used to perform the substitution.
   * Recursive traversal of the expression tree
   * is provided by {@link ExprTransformer}.
   */
  private final class Worker extends ExprTransformer {
    /** Number of times each formula variable is bound. */
    private final Map<String, Integer> boundFormulaVars;
    
    /** Number of times each term variable is bound. */;
    private final Map<String, Integer> boundTermVars;
    
    /** Constructor. */
    Worker() {
      super(new SimpleExpressionFactory());
      boundTermVars = new HashMap<String, Integer>();
      boundFormulaVars = new HashMap<String, Integer>();
    }

    /**
     * Check if a term variable is bound at least once.
     * @param var Variable name.
     * @return True if bound, false otherwise.
     */
    private boolean isTermVariableBound(final String var) {
      return boundTermVars.containsKey(var) && 
             boundTermVars.get(var) == 0;
    }
    
    /**
     * Check if a formula variable is bound at least once.
     * @param var Variable name.
     * @return True if bound, false otherwise.
     */
    private boolean isFormulaVariableBound(final String var) {
      return boundFormulaVars.containsKey(var) && 
             boundFormulaVars.get(var) == 0;
    }
    
    /**
     * Modify an integer stored in a map under given key.
     * If the key is not bound, the value is initialized to zero
     * before the change is applied.
     * @param map Map with values.
     * @param key Key.
     * @param change Change amount.
     */
    private void modify(final Map<String, Integer> map,
                        final String key,
                        final int change) {
      if (!map.containsKey(key)) {
        map.put(key, 0);
      }
      map.put(key, map.get(key) + change);
    }
    
    /** {@inheritDoc} */
    @Override
    public Term caseVariable(final String var) {
      if (!isTermVariableBound(var)) {
        if (termMap.containsKey(var)) {
          return termMap.get(var);
        }
      }
      return getFactory().createVariable(var);
    }

    /** {@inheritDoc} */
    @Override
    public Term caseLet(final String var, 
                        final ValueType type, 
                        final Term value, 
                        final Term body) {
      modify(boundTermVars, var, 1);
      final Term result = super.caseLet(var, type, value, body);
      modify(boundTermVars, var, -1);
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public Term caseLetProp(final String var, 
                            final Formula value, 
                            final Term body) {
      modify(boundFormulaVars, var, 1);
      final Term result = super.caseLetProp(var, value, body);
      modify(boundFormulaVars, var, -1);
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public Formula caseFormulaVariable(final String var) {
      if (!isFormulaVariableBound(var)) {
        if (termMap.containsKey(var)) {
          return formulaMap.get(var);
        }
      }
      return getFactory().createFormulaVariable(var);
    }

    /** {@inheritDoc} */
    @Override
    public Formula caseExists(final Iterable<VariableBinding> variables,
                              final Formula arg,
                              final Expression... patterns) {
      for (final VariableBinding variable : variables) {
        modify(boundTermVars, variable.getName(), 1);
      }
      final Formula result = super.caseExists(variables, arg, patterns);
      for (final VariableBinding variable : variables) {
        modify(boundTermVars, variable.getName(), -1);
      }
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public Formula caseForall(final Iterable<VariableBinding> variables,
                              final Formula arg,
                              final Expression... patterns) {
      for (final VariableBinding variable : variables) {
        modify(boundTermVars, variable.getName(), 1);
      }
      final Formula result = super.caseForall(variables, arg, patterns);
      for (final VariableBinding variable : variables) {
        modify(boundTermVars, variable.getName(), -1);
      }
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public Formula caseFLet(final String var, 
                            final ValueType type, 
                            final Term value, 
                            final Formula body) {
      modify(boundTermVars, var, 1);
      final Formula result = super.caseFLet(var, type, value, body);
      modify(boundTermVars, var, -1);
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public Formula caseFLetProp(final String var, 
                                final Formula value, 
                                final Formula body) {
      modify(boundFormulaVars, var, 1);
      final Formula result = super.caseFLetProp(var, value, body);
      modify(boundFormulaVars, var, -1);
      return result;
    }
  }
}
