package net.afyre.haha.demo.ui.handler;


import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Scanner;

import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleConstants;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.IConsoleView;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;
import org.eclipse.ui.handlers.HandlerUtil;
import org.eclipse.ui.services.ISourceProviderService;
import org.eclipse.xtext.nodemodel.INode;
import org.eclipse.xtext.nodemodel.util.NodeModelUtils;
import org.eclipse.xtext.resource.XtextResource;
import org.eclipse.xtext.ui.editor.XtextEditor;
import org.eclipse.xtext.ui.editor.model.IXtextDocument;
import org.eclipse.xtext.util.concurrent.IUnitOfWork;

import net.afyre.haha.demo.coq.CoqLanguage;
import net.afyre.haha.demo.coq.VCCoqPrinter;
import net.afyre.haha.demo.haha.Program;
import net.afyre.haha.demo.solver.SolverRunningState;
import net.afyre.haha.demo.ui.HahaSourceProvider;
import net.afyre.haha.demo.ui.annotations.UnknownAssertAnnotation;
import net.afyre.haha.demo.ui.internal.HahaActivator;
import net.afyre.haha.demo.ui.prefs.CoqPreferencePage;
import net.afyre.haha.demo.util.CmpUtil;
import net.afyre.haha.demo.util.FileFinder;
import net.afyre.haha.demo.vcgen.HoareVCGen;
import net.afyre.haha.demo.vcgen.VCCollector;
import net.afyre.haha.demo.vcgen.VCContext;
import net.afyre.haha.demo.vcgen.VCPrinter;
import net.afyre.haha.demo.vcgen.VerificationCondition;

/**
 * The handler of the buttons to generate Coq verification conditions.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @author Aleksy Schubert <alx@mimuw.edu.pl>
 */
public class CoqVCGenHandler extends HahaAbstractHandler implements IHandler {
  
  private static final String HOWTO_FILE_NAME = "resources/howto.txt";

  private static final String CONSOLE_NAME = "VerificationConditions";

  private FileFinder fileFinder;
  
  private HashSet<String> checkedAxioms;
  
  private int line;

  private IWorkbenchPage mpage;

  /** {@inheritDoc} */
  @Override
  public Object execute(final ExecutionEvent event) throws ExecutionException {
    if (fileFinder == null) {
      fileFinder = new FileFinder();
    }
    final ISourceProviderService sourceProviderService = 
        (ISourceProviderService) HandlerUtil
          .getActiveWorkbenchWindow(event)
          .getService(ISourceProviderService.class);
    final HahaSourceProvider sourceProvider = 
        (HahaSourceProvider) sourceProviderService
          .getSourceProvider(HahaSourceProvider.VAR_RUNNING);

    final Object editorObject = HandlerUtil.getVariable(event, "activeEditor");
    if (!(editorObject instanceof XtextEditor)) {
      final MessageBox box = new MessageBox(PlatformUI.getWorkbench()
          .getActiveWorkbenchWindow().getShell());
      box.setMessage("Oops: " + editorObject.getClass().getCanonicalName());
      box.open();
      return null;
    }
    actOnEditor(event, sourceProvider, editorObject);
    return null;
  }

  private void actOnEditor(final ExecutionEvent event,
                           final HahaSourceProvider sourceProvider,
                           final Object editorObject) {
    final XtextEditor editor = (XtextEditor) editorObject;
    final IXtextDocument doc = editor.getDocument();
    final IResource editorResource = editor.getResource();
    final IAnnotationModel annotationModel = editor.getDocumentProvider()
        .getAnnotationModel(editor.getEditorInput());
    final Thread thread = new Thread() {
      @Override
      public void run() {
        doc.readOnly(new IUnitOfWork<Object, XtextResource>() {
          public Object exec(final XtextResource resource) {
            sourceProvider.setRunning(true);
            try {
              final Program prog = (Program) resource.getContents().get(0);
              runVCGen(editorResource, annotationModel, prog, HandlerUtil
                  .getActiveWorkbenchWindow(event).getActivePage());
            } finally {
              sourceProvider.setRunning(false);
            }
            return null;
          }
        });
      }
    };
    sourceProvider.setThread(thread);
    thread.start();
  }

  /**
   * The method prints a particular message to the given writer
   * with handling some special cases.
   * 
   * @param w the writer to write to
   * @param msg the message to write
   */
  private void print(final Writer w, final String msg) {
    final String message = msg == null ? "Null" : msg;
    try {
      w.write("CoqVCGen> " + message);
      w.flush();
    } catch (final IOException ex) {
      // Ignore.
    }
  }

  private void runVCGen(final IResource resource, 
                        final IAnnotationModel annotationModel,
                        final Program program, final IWorkbenchPage page) {
    final HoareVCGen vcgen = new HoareVCGen();

    final MessageConsole console = findConsole(CONSOLE_NAME);
    final MessageConsoleStream out = console.newMessageStream();
    final Writer writer = new OutputStreamWriter(out);

    print(writer, "Line number: " + getLineNo(page) + "\n");

    final IPreferenceStore store = HahaActivator.getInstance()
        .getPreferenceStore();
    final String coqFilePath = store
        .getDefaultString(CoqPreferencePage.PREF_COQ_FILE_PATH);
    final String mainFileName = getMainFileName(page);

    final FileOutputStream coqout =
        getCoqFileOutput(mainFileName, console, writer, coqFilePath);
    if (coqout == null) {
      showConsole(page, console);
      return;
    }


    final VCCollector collector = new VCCollector();
    final Writer coqwriter = new OutputStreamWriter(coqout);
    final VCPrinter printer = new VCPrinter(writer);
    final VCCoqPrinter coqprinter = new VCCoqPrinter(coqwriter);

    print(writer, "====================================\n");
    print(writer, " Collecting verification conditions\n");
    print(writer, "====================================\n\n");
    try {
      vcgen.generateVerificationConditions(null, program, collector);
    } catch (final Exception ex) {
      ex.printStackTrace(new PrintWriter(writer));
      try {
        writer.flush();
      } catch (final IOException e2) {
      }
    }
    print(writer, "Verification conditions computed.\n");
    print(writer, "Conditions: ");

    for (final VerificationCondition cond : collector.getConditions()) {
      printer.processVerificationCondition(cond);
      coqprinter.processVerificationCondition(cond);
    }

    printer.end();
    clearMarkers(resource, annotationModel);
    coqprinter.end();
    try {
      coqprinter.addHowto(FileLocator
          .openStream(HahaActivator.getInstance().getBundle(),
                      new Path(HOWTO_FILE_NAME), false));
    } catch (final IOException e3) {
      print(writer,
            "Cannot append HOWTO instructions. " +
            "We try to prove conditions anyway.");
    }
    try {
      coqwriter.close();
      coqout.close();
    } catch (final IOException e1) {
      e1.printStackTrace(new PrintWriter(writer));
      try {
        writer.flush();
      } catch (final IOException e2) {
      }
      showConsole(page, console);
      return;
    }

    print(writer, "\n\n");

    print(writer, "================\n");
    print(writer, " Running CoqIDE\n");
    print(writer, "================\n\n");

    try {
      writer.flush();
    } catch (final IOException e) {
    }

    showConsole(page, console);
    if (callCoqIDE(store, coqFilePath, mainFileName, page, writer)) {
      print(writer, "================\n");
      print(writer, " CoqIDE finished\n");
      print(writer, "================\n\n");
      try {
        writer.flush();
      } catch (final IOException e) {
        //Ignore
      }
      print(writer, "================\n");
      print(writer, " Running Coq compilation\n");
      print(writer, "================\n\n");
      if (callCoqc(store, coqFilePath, mainFileName, page, writer)) {
        print(writer, "================\n");
        print(writer, " Checking Coq conditions\n");
        print(writer, "================\n\n");      
        final HashSet<String> axioms = callCoqcheck(store, coqFilePath, 
                                                    mainFileName,
                                                    page, writer);
        try {
          writer.flush();
        } catch (final IOException e) {
        }
        print(writer, String.format("Total number of conditions: %d\n",
                                    collector.getConditions().size()));
        markErrors(axioms, coqprinter.getVCs(), resource, annotationModel,
                   writer);
      }
    }
  }

  private void markErrors(final HashSet<String> axioms,
                          final 
                            Hashtable<String, VerificationCondition> hashtable,
                          final IResource resource, 
                          final IAnnotationModel annotationModel,
                          final Writer writer) {
    int unknown = 0;
    int valid = 0;
    final Enumeration<String> e = hashtable.keys();

    while (e.hasMoreElements()) {
      final String string = e.nextElement();
      if (axioms.contains(string)) {
        createUnknownMarkers(resource, annotationModel, hashtable.get(string));
        unknown++;
      } else {
        valid++;
      }
    }
    print(writer, String.format("Valid: %d\n", valid));
    print(writer, String.format("Unknown: %d\n", unknown));
  }

  private HashSet<String> callCoqcheck(final IPreferenceStore store,
                                       final String coqFilePath, 
                                       final String mainFileName,
                                       final IWorkbenchPage page,
                                       final Writer writer) {
    // Build the command
    final String[] cmd = {null, "-o", "-admit", "ZArith", 
                          "vcs_" + mainFileName};
    cmd[0] = store.getString(CoqPreferencePage.PREF_COQ_CHECKER_PATH);

    final File directory = new File(coqFilePath);
    final Process pr = runCommand(directory, cmd);
    SolverRunningState.addActiveVerificationProcess(pr);
    parseCheck(pr.getInputStream());
    try {
      final int res = pr.waitFor();
      if (res != 0) {
        showError("Cannot check VCs from " + mainFileName + ".v", page,
                  writer);
      } else {
        showError("VC check OK", page, writer);
      }
    } catch (final InterruptedException ex) {
      // TODO Auto-generated catch block
      ex.printStackTrace();
    }
    SolverRunningState.rmActiveVerificationProcess(pr);
    return checkedAxioms;
  }

  private enum States {
    INIT, AXIOMS
  }

  private void parseCheck(final InputStream inputStream) {
    final BufferedInputStream bs = new BufferedInputStream(inputStream);
    final Scanner scanner = new Scanner(bs);
    States state = States.INIT;
    while (scanner.hasNextLine()) {
      final StringBuffer iline = new StringBuffer(scanner.nextLine().trim());
      state = interpretLine(state, iline);
    }
    scanner.close();
    try {
      inputStream.close();
    } catch (final IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private States interpretLine(final States state, final StringBuffer iline) {
    States res = state;
    switch (state) {
      case INIT:
        if (iline.toString().compareTo("* Axioms:") == 0) {
          checkedAxioms = new HashSet<String>();
          res = States.AXIOMS;
        } else {
          res = state;
        }
        break;
      case AXIOMS:
        final String item = iline.substring(iline.indexOf(".") + 1).toString();
        checkedAxioms.add(item);
        res = state;
        break;
      default:
        res = state;
        break;
    }
    return res;
  }

  /**
   * This method calls coqc program with a file that contains the
   * verification conditions.
   * 
   * @param store        store object to retrieve the file name 
   *                     of coqc
   * @param coqFilePath  the path where the Coq file is located
   * @param mainFileName the basename of the Coq file to call coqide for
   * @param page         the workbench page to make possible error display
   * @param writer 
   * @return true in case the compilation went correctly
   */
  private boolean callCoqc(final IPreferenceStore store, 
                           final String coqFilePath,
                           final String mainFileName, 
                           final IWorkbenchPage page,
                           final Writer writer) {
    // Build the command
    final String[] cmd = new String[2];
    cmd[0] = store.getString(CoqPreferencePage.PREF_COQ_SOLVER_PATH);

    cmd[1] = getCoqFileName(coqFilePath, mainFileName) + ".v";
    final File directory = new File(coqFilePath);
    final Process pr = runCommand(directory, cmd);
    SolverRunningState.addActiveVerificationProcess(pr);
    try {
      final int res = pr.waitFor();
      SolverRunningState.rmActiveVerificationProcess(pr);
      if (res != 0) {
        showError("Cannot compile VCs in file vcs_" + mainFileName + ".v", page,
                  writer);
        return false;
      } else {
        showError("VC compilation OK", page, writer);
        return true;
      }
    } catch (final InterruptedException ex) {
      // TODO Auto-generated catch block
      SolverRunningState.rmActiveVerificationProcess(pr);
      ex.printStackTrace();
    }
    return false;
  }

  /**
   * This method calls coqide program with a file that contains the
   * verification conditions.
   * 
   * @param store        store object to retrieve the file name 
   *                     of coqide
   * @param coqFilePath  the path where the Coq file is located
   * @param mainFileName the basename of the Coq file to call coqide for
   * @param page         the workbench page to make possible error display
   * @param writer 
   * @return 
   */
  private boolean callCoqIDE(final IPreferenceStore store,
                          final String coqFilePath,
                          final String mainFileName,
                          final IWorkbenchPage page,
                          final Writer writer) {
    // Fetch the file with tactics.
    bringTacticsFile(coqFilePath);

    // Build the command
    final String[] cmd = new String[2];
    cmd[0] = store.getString(CoqPreferencePage.PREF_COQ_IDE_PATH);

    cmd[1] = getCoqFileName(coqFilePath, mainFileName) + ".v";
    final File directory = new File(coqFilePath);
    final Process pr = runCommand(directory, cmd);
    SolverRunningState.addActiveVerificationProcess(pr);
    try {
      pr.waitFor();
    } catch (final InterruptedException ex) {
      showError("CoqIDE process stopped.", page, writer);
      SolverRunningState.rmActiveVerificationProcess(pr);
      return false;
    }
    if (pr.exitValue() != 0) {
      showError("CoqIDE process returned wrongly.", page, writer);
      SolverRunningState.rmActiveVerificationProcess(pr);
      return false;
    }
    SolverRunningState.rmActiveVerificationProcess(pr);
    return true;
  }

  private String getCoqFileName(final String coqFilePath,
                                final String mainFileName) {
    return coqFilePath + File.separator + "vcs_" +
           CoqLanguage.coqifyName(mainFileName);
  }

  private Process runCommand(final File directory, final String[] cmd) {
    final ProcessBuilder processBuilder = new ProcessBuilder(cmd);
    processBuilder.directory(directory);
    try {
      return processBuilder.start();
    } catch (final IOException ex) {
      // TODO Auto-generated catch block
      ex.printStackTrace();
    }
    return null;
  }

  private void bringTacticsFile(final String coqFilePath) {
    final InputStream stream;
    try {
      stream = FileLocator.openStream(HahaActivator.getInstance().getBundle(),
                                      new Path("resources/tactics.v"), false);
      final File dest = new File(coqFilePath + File.separator + "tactics.v");

      Files.copy(stream, dest.toPath());

    } catch (final FileAlreadyExistsException e0) {
      //We silently ignore it.
    } catch (final IOException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    }
  }

  private FileOutputStream getCoqFileOutput(final String mainFileName,
                                            final MessageConsole console,
                                            final Writer writer,
                                            final String coqFilePath) {
    FileOutputStream coqout = null;
    final String fname = getCoqFileName(coqFilePath, mainFileName);
    renameIfExistsCoq(fname);
    try {
      coqout = new FileOutputStream(fname + ".v");
    } catch (final FileNotFoundException ex) {
      print(writer, "Coq file: " + fname + ".v" + " cannot be created.\n");
      print(writer, "Check if the directory exists.\n");
      return null;
    }
    print(writer, "Coq file: " + fname + ".v" + "\n");
    return coqout;
  }

  /**
   * The procedure tries to rename the current file under fname.v
   * name to the first available one of the name fname<num>.v.
   * The method does not copy the current file in case the last
   * file of the name fname<num>.v has the same contents as the current
   * file under fname.v
   * 
   * @param fname the basename of a file to rename
   */
  private void renameIfExistsCoq(final String fname) {
    final String namev = fname + ".v";
    File f = fileFinder.findFileOnPath(namev);
    int num = 0;
    while (f.exists()) {
      num++;
      f = fileFinder.findFileOnPath(fname + num + ".v");
    }
    if (num >= 1) {
      InputStream source = null;
      try {
        source = new FileInputStream(namev);
      } catch (final FileNotFoundException e1) {
        // We silently ignore the impossible exception.
      }
      if (num == 1 || CmpUtil.differFiles(namev, fname + (num - 1) + ".v")) {
        final File dest = new File(fname + num + ".v");
        try {
          Files.copy(source, dest.toPath());
        } catch (final IOException ex) {
          // TODO Auto-generated catch block
          ex.printStackTrace();
        }
      }
    }
  }



  private int getLineNo(final IWorkbenchPage page) {
    this.mpage = page;

    Display.getDefault().syncExec(new Runnable() {

      public void run() {
        final IEditorPart editorPart = mpage.getActiveEditor();
        final ITextSelection textSelection = 
            (ITextSelection) editorPart.getSite()
              .getSelectionProvider().getSelection();
        line = textSelection.getStartLine();
      }
    });

    return line;
  }

  /**
   * Returns the main part of the file name edited in the given page,
   * omitting its extension.
   * 
   * @param page the page to look for the file name in
   * @return the resulting name
   */
  private String getMainFileName(final IWorkbenchPage page) {
    final String fullname = page.getActiveEditor().getEditorInput().getName();
    final int endpos = fullname.lastIndexOf('.');
    return fullname.substring(0, endpos);
  }

  private void createUnknownMarkers(final IResource resource,
                                    final IAnnotationModel annotationModel,
                                    final VerificationCondition vc) {
    if (resource != null) {
      for (final EObject srcElement : vc.getGoal().getSource()) {
        final INode node = NodeModelUtils.findActualNodeFor(srcElement);

        try {
          final IMarker marker = resource
              .createMarker("net.afyre.haha.demo.ui.problemmarker");
          String msg = "Unable to prove the verification condition.\n";
          msg += getVCDescription(vc);
          marker.setAttribute(IMarker.MESSAGE, msg);
          marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
          marker.setAttribute(IMarker.LINE_NUMBER, node.getStartLine());
          marker.setAttribute(IMarker.CHAR_START, node.getOffset());
          marker.setAttribute(IMarker.CHAR_END, node.getTotalEndOffset());
        } catch (final CoreException ex) {
          // TODO Auto-generated catch block
          ex.printStackTrace();
        }
      }
    } else {
      for (final EObject srcElement : vc.getGoal().getSource()) {
        final INode node = NodeModelUtils.findActualNodeFor(srcElement);
        final Position position = new Position(node.getOffset(),
                                               node.getLength());
        final Annotation annotation = new UnknownAssertAnnotation(vc);
        annotationModel.addAnnotation(annotation, position);
      }
    }
  }

  private String getVCDescription(final VerificationCondition vc) {
    final List<String> contexts = new ArrayList<String>();

    VCContext current = vc.getContext();
    while (current != null) {
      contexts.add(current.getComment() + "\n");
      current = current.getParent();
    }
    final StringBuilder result = new StringBuilder();
    for (int i = contexts.size() - 1; i >= 0; i--) {
      result.append(contexts.get(i));
    }
    return result.toString();
  }

  private void showConsole(final IWorkbenchPage page,
                           final MessageConsole console) {
    final String id = IConsoleConstants.ID_CONSOLE_VIEW;
    final Display display = page.getWorkbenchWindow().getShell().getDisplay();

    display.syncExec(new Runnable() {
      @Override
      public void run() {
        final IConsoleView view;
        try {
          view = (IConsoleView) page.showView(id);
          view.display(console);
        } catch (final PartInitException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    });
  }

  private MessageConsole findConsole(final String name) {
    final ConsolePlugin plugin = ConsolePlugin.getDefault();
    final IConsoleManager conMan = plugin.getConsoleManager();
    final IConsole[] existing = conMan.getConsoles();
    for (int i = 0; i < existing.length; i++) {
      if (name.equals(existing[i].getName())) {
        return (MessageConsole) existing[i];
      }
    }
    //no console found, so create a new one
    final MessageConsole myConsole = new MessageConsole(name, null);
    conMan.addConsoles(new IConsole[] {myConsole });
    return myConsole;
  }

  private void showError(final String message, final IWorkbenchPage page,
                         final Writer writer) {
    print(writer, "Remark: " + message + "\n");
  }
}
