package net.afyre.haha.demo.util.tests;

import static net.afyre.haha.demo.util.tests.Util.*;
import static org.junit.Assert.*;

import java.util.Iterator;

import net.afyre.haha.demo.util.ArrayIterable;
import net.afyre.haha.demo.util.ArrayIterator;

import org.junit.Test;

/**
 * Tests for the {@link ArrayIterable} class.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @see ArrayIterator
 */
public class ArrayIterableTest {
  /** GetInstance does not return null. */
  @Test
  public void testInstanceNotNull() {
    final Integer[] array = { 1, 2, 3, 4 };
    Iterable<Integer> iterable;
    
    iterable = ArrayIterable.getInstance(array);
    assertNotNull("Null returned from getInstance/1", iterable);
    iterable = ArrayIterable.getInstance(array, 0);
    assertNotNull("Null returned from getInstance/2", iterable);
    iterable = ArrayIterable.getInstance(array, 0, 0);
    assertNotNull("Null returned from getInstance/3", iterable);
  }
  
  /** NPE is thrown if null is passed to getInstance()/1 */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceWithNull1() {
    ArrayIterable.getInstance(null);
  }
  
  /** NPE is thrown if null is passed to getInstance()/2 */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceWithNull2() {
    ArrayIterable.getInstance(null, 0);
  }
  
  /** NPE is thrown if null is passed to getInstance()/3 */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceWithNull3() {
    ArrayIterable.getInstance(null, 0, 0);
  }
  
  /** All elements are returned when iterating over a whole array. */
  @Test
  public void testSimpleIteration() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterable<Integer> iterable = ArrayIterable.getInstance(array);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** All elements are returned when iterating with offset = 0. */
  @Test
  public void testSimpleIterationWithZeroOffset1() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterable<Integer> iterable = ArrayIterable.getInstance(array, 0);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** All N elements are returned for offset = 0 and bound = N-1. */
  @Test
  public void testSimpleIterationWithZeroOffset2() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterable<Integer> iterable = ArrayIterable.getInstance(array, 0, 3);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** Correct elements are returned when iterating with an offset. */
  @Test
  public void testSimpleIterationWithOffset() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterable<Integer> iterable = 
        ArrayIterable.getInstance(array, 2);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it, 3, 4, 5, 6);
  }
  
  /** Correct elements are returned when iterating with offset + bound. */
  @Test
  public void testSimpleIterationWithOffsetAndBound() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterable<Integer> iterable = 
        ArrayIterable.getInstance(array, 2, 4);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it, 3, 4, 5);
  }
    
  /** Offset greater than array length does not cause an error. */
  @Test
  public void testOffsetGtArrayLength() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    Iterable<Integer> iterable;
    Iterator<Integer> it;

    iterable = ArrayIterable.getInstance(array, 6);
    it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it);

    iterable = ArrayIterable.getInstance(array, 6);
    it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it);
  }
  
  /** Bound greater than array length does not cause problems. */
  @Test
  public void TestBoundGtLength() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterable<Integer> iterable = 
        ArrayIterable.getInstance(array, 2, 1024);
    final Iterator<Integer> it = iterable.iterator();
    assertIteratorReturns(it, 3, 4, 5, 6);    
  }
  
  /** Offset greater than bound does not cause an error. */
  @Test
  public void testOffsetGtBound() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    Iterable<Integer> iterable;
    Iterator<Integer> it;

    iterable = ArrayIterable.getInstance(array, 4, 3);
    it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it);

    iterable = ArrayIterable.getInstance(array, 7, 3);
    it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it);
  }
  
  /** Empty array does not cause an error. */
  @Test
  public void testEmptyArray() {
    final Integer[] array = { };
    final Iterable<Integer> iterable = ArrayIterable.getInstance(array);
    final Iterator<Integer> it = iterable.iterator();
    assertNotNull("Null iterator returned.", it);
    assertIteratorReturns(it);
  }
}
