package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;
import static net.afyre.haha.demo.util.tests.Util.*;

import java.util.Iterator;

import net.afyre.haha.demo.util.ArrayIterator;

import org.junit.Test;

/**
 * Tests for the {@link ArrayIterator} class.
 * Test cases defined here are usually duplicated
 * in ArrayIterableTest.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @see ArrayIterableTest
 */
public class ArrayIteratorTest {

  /** GetInstance does not return null. */
	@Test
	public void testInstanceNotNull() {
		final Integer[] array = { 1, 2, 3, 4 };
		Iterator<Integer> it;
		
		it = ArrayIterator.getInstance(array);
		assertNotNull("Null returned from getInstance/1", it);
    it = ArrayIterator.getInstance(array, 0);
    assertNotNull("Null returned from getInstance/2", it);
    it = ArrayIterator.getInstance(array, 0, 0);
    assertNotNull("Null returned from getInstance/3", it);
	}

	/** NPE is thrown if null is passed to getInstance()/1 */
	@Test(expected=NullPointerException.class)
	public void testGetInstanceWithNull1() {
	  ArrayIterator.getInstance(null);
	}
	
  /** NPE is thrown if null is passed to getInstance()/2 */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceWithNull2() {
    ArrayIterator.getInstance(null, 0);
  }
	
  /** NPE is thrown if null is passed to getInstance()/3 */
  @Test(expected=NullPointerException.class)
  public void testGetInstanceWithNull3() {
    ArrayIterator.getInstance(null, 0, 0);
  }
  
  /** All elements are returned when iterating over a whole array. */
  @Test
  public void testSimpleIteration() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** All elements are returned when iterating with offset = 0. */
  @Test
  public void testSimpleIterationWithZeroOffset1() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array, 0);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** All N elements are returned for offset = 0 and bound = N-1. */
  @Test
  public void testSimpleIterationWithZeroOffset2() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array, 0, 3);
    assertIteratorReturns(it, 1, 2, 3, 4);
  }
  
  /** Correct elements are returned when iterating with an offset. */
  @Test
  public void testSimpleIterationWithOffset() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array, 2);
    assertIteratorReturns(it, 3, 4, 5, 6);
  }
  
  /** Correct elements are returned when iterating with offset + bound. */
  @Test
  public void testSimpleIterationWithOffsetAndBound() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array, 2, 4);
    assertIteratorReturns(it, 3, 4, 5);
  }
    
  /** Offset greater than array length does not cause an error. */
  @Test
  public void testOffsetGtArrayLength() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    Iterator<Integer> it;
    
    it = ArrayIterator.getInstance(array, 6);
    assertIteratorReturns(it);
    it = ArrayIterator.getInstance(array, 7);
    assertIteratorReturns(it);
  }
  
  /** Offset greater than bound does not cause an error. */
  @Test
  public void testOffsetGtBound() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    Iterator<Integer> it;
    
    it = ArrayIterator.getInstance(array, 4, 3);
    assertIteratorReturns(it);
    it = ArrayIterator.getInstance(array, 7, 3);
    assertIteratorReturns(it);
  }
  
  /** Bound greater than array length does not cause problems. */
  @Test
  public void TestBoundGtLength() {
    final Integer[] array = { 1, 2, 3, 4, 5, 6 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array, 2, 100);
    assertIteratorReturns(it, 3, 4, 5, 6);    
  }
  
  /** Empty array does not cause an error. */
  @Test
  public void testEmptyArray() {
    final Integer[] array = { };
    final Iterator<Integer> it = ArrayIterator.getInstance(array);
    assertIteratorReturns(it);
  }
  
  /** Exception thrown when calling remove(). */
  @Test(expected=UnsupportedOperationException.class)
  public void testRemove() {
    final Integer[] array = { 1, 2, 3, 4 };
    final Iterator<Integer> it = ArrayIterator.getInstance(array);
    it.remove();
  }
}
