package net.afyre.haha.demo.util;

import java.util.Iterator;

/**
 * An iterable wrapper which discards given number
 * of initial elements.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 * @param <T> Type of elements.
 */
public class OffsetIterable<T> implements Iterable<T> {
  /** Wrapped iterable. */
  private final Iterable<? extends T> source;
  
  /** Number of elements to be skipped. */
  private final int offset;
  
  /**
   * Constructor.
   * @param source Source iterable.
   * @param offset Number of elements to be skipped.
   *               This value may be greater than
   *               the number of elements returned by 
   *               the source iterator.
   * @throws NullPointerException If the source iterator is null.
   * @throws IllegalArgumentException If the offset is negative.
   */
  private OffsetIterable(final Iterable<? extends T> source,
                         final int offset) {
    if (source == null) {
      throw new NullPointerException("source");
    }
    if (offset < 0) {
      throw new IllegalArgumentException("offset");
    }
    this.source = source;
    this.offset = offset;
  }

  /**
   * Create an instance of this class.
   * @param source Source iterable.
   * @param offset Number of elements to be skipped.
   *               This value may be greater than
   *               the number of elements returned by 
   *               the source iterator.
   * @throws NullPointerException If the source iterator is null.
   * @throws IllegalArgumentException If the offset is negative.
   */
  public static<T> OffsetIterable<T> getInstance(
      final Iterable<? extends T> source,
      final int offset) {
    return new OffsetIterable<T>(source, offset);
  }
  
  /** {@inheritDoc} */
  @Override
  public Iterator<T> iterator() {
    final Iterator<? extends T> result = source.iterator();
    for (int i = 0; i < offset && result.hasNext(); i++) {
      result.next();
    }
    return Wrapper.getInstance(result);
  }
  
  /** 
   * We must wrap the source iterator to ensure 
   * proper behavior of remove(). 
   * @param <T> Type of returned elements.
   **/
  private static final class Wrapper<T> implements Iterator<T> {
    /** Wrapped iterator. */
    private final Iterator<? extends T> source;
    
    /** Flag set after an element is returned by next() */
    private boolean firstElementReturned;
    
    /** 
     * Constructor. 
     * @param source Wrapped iterator. 
     * @throws NullPointerException If the source iterator is null.
     **/
    private Wrapper(final Iterator<? extends T> source) {
//      if (source == null) {
//        throw new NullPointerException("source");
//      }
      this.source = source;
      this.firstElementReturned = false;
    }
    
    /**
     * Wrap an iterator.
     * 
     * @param <T> Type of returned elements.
     * @param source Wrapped iterator. 
     * @return New iterator.
     * @throws NullPointerException If the source iterator is null.
     **/
    public static<T> Iterator<T> getInstance(
        final Iterator<? extends T> source) {
      return new Wrapper<T>(source);
    }
    
    /** {@inheritDoc} */
    @Override
    public boolean hasNext() {
      return source.hasNext();
    }

    /** {@inheritDoc} */
    @Override
    public T next() {
      final T result = source.next();
      // Not set if next() throws an exception.
      firstElementReturned = true;
      return result;
    }

    /** {@inheritDoc} */
    @Override
    public void remove() {
      if (firstElementReturned) {
        source.remove();
      } else {
        throw new IllegalStateException("Remove() before next()");
      }
    }
  }
}
