package net.afyre.haha.demo.util.async;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import net.afyre.haha.demo.util.Function;

/**
 * Applies a function to a result of an asynchronous computation.
 * 
 * Note that the function is executed synchronously, 
 * so it must not be blocking.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * 
 * @param <FirstResult> Type of the result of the first computation.
 * @param <Result> Type of the final result.
 */
public class MappedFuture<FirstResult, Result> implements Future<Result> {
  /** First computation. */
  private final Future<? extends FirstResult> future;
  
  /** Mapped function. */
  private final Function<? super FirstResult, ? extends Result> function;
  
  /**
   * Constructor.
   * @param <FirstResult> Type of the result of the first computation.
   * @param <Result> Type of the final result.
   * @param future Asynchronous part of the computation.
   * @param function Synchronous part of the computation.
   */
  public static<FirstResult, Result> MappedFuture<FirstResult, Result>
  getInstance(final Future<? extends FirstResult> future,
              final Function<? super FirstResult, 
                             ? extends Result> function) {
    return new MappedFuture<FirstResult, Result>(future, function);
  }
  
  /**
   * Constructor.
   * @param future Asynchronous part of the computation.
   * @param function Synchronous part of the computation.
   */
  protected MappedFuture(final Future<? extends FirstResult> future,
                         final Function<? super FirstResult, 
                                        ? extends Result> function) {
    this.future = future;
    this.function = function;
  }

  /** {@inheritDoc} */
  @Override
  public boolean cancel(boolean mayInterruptIfRunning) {
    return future.cancel(mayInterruptIfRunning);
  }

  /** {@inheritDoc} */
  @Override
  public Result get() throws InterruptedException, ExecutionException {
    return function.apply(future.get());
  }

  /** {@inheritDoc} */
  @Override
  public Result get(final long timeout, final TimeUnit unit) 
    throws InterruptedException, ExecutionException, TimeoutException {
    return function.apply(future.get(timeout, unit));
  }

  /** {@inheritDoc} */
  @Override
  public boolean isCancelled() {
    return future.isCancelled();
  }

  /** {@inheritDoc} */
  @Override
  public boolean isDone() {
    return future.isDone();
  }
}
