package net.afyre.haha.demo.vcgen;

import java.util.List;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

import net.afyre.haha.demo.haha.Assignment;
import net.afyre.haha.demo.haha.Expression;
import net.afyre.haha.demo.haha.HahaPackage;
import net.afyre.haha.demo.haha.SimpleStatement;
import net.afyre.haha.demo.haha.Skip;
import net.afyre.haha.demo.solver.expr.util.Substitution;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.terms.impl.ArrayAccess;
import net.afyre.haha.demo.solver.expr.terms.impl.ArrayUpdate;
import net.afyre.haha.demo.solver.expr.terms.impl.Variable;


/**
 * VCGen implementation for simple statements.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class DefaultSimpleStatementVCGen 
    implements SimpleStatementVCGen {
  
  /** Object used to convert AST expressions to terms. */
  private final ExpressionUtil exprUtil = ExpressionUtil.getInstance();

  /** Utility methods. */
  private final VCGenUtil util = VCGenUtil.getInstance();
  
  /** {@inheritDoc} */
  public void generateVerificationConditions(
      final VCContext context,                                             
      final List<DocumentedFormula> preconditions,
      final SimpleStatement stmt,
      final List<DocumentedFormula> postconditions,
      final VCConsumer consumer) {
    
    // Generate one VC per postcondition.
    for (final DocumentedFormula postcondition : postconditions) {
      processStatement(context, 
                       preconditions, stmt, postcondition, 
                       consumer);
    }
  }

  /**
   * Process a statement.
   * @param context Context information for created conditions.
   * @param preconditions Preconditions.
   * @param stmt Statement.
   * @param postconditon Postcondition.
   * @param consumer Object used process generated conditions.
   */
  private void processStatement(
      final VCContext context,
      final List<DocumentedFormula> preconditions,
      final SimpleStatement stmt,
      final DocumentedFormula postcondition,
      final VCConsumer consumer) {
    final VCContext processedContext = 
        util.processFunctionCalls(context, preconditions, stmt, consumer);
    switch (stmt.eClass().getClassifierID()) {
    case HahaPackage.ASSIGNMENT:
      processAssignment(processedContext,
                        preconditions, 
                        (Assignment)stmt, 
                        postcondition,
                        consumer);
      break;
    case HahaPackage.SKIP:
      processSkip(processedContext,
                  preconditions, (Skip)stmt, postcondition,
                  consumer);
    default: 
      assert(false);
      break;
    }    
  }
  
  /**
   * Process an assignment statement.
   * @param context Context information for created conditions.
   * @param preconditions Preconditions.
   * @param stmt Statement.
   * @param postconditon Postcondition.
   * @param consumer Object used process generated conditions.
   */
  private void processAssignment(
      final VCContext context,
      final Iterable<DocumentedFormula> preconditions,
      final Assignment stmt,
      final DocumentedFormula postcondition,
      final VCConsumer consumer) {

    // TODO: More robust naming scheme
    final String varName = stmt.getTarget().getVariable().getName();
    
    final Variable target = new Variable(varName);
    Term value = exprUtil.translateTerm(stmt.getValue());
    
    Term array = target;
    final List<Expression> reversedPath = 
        Lists.reverse(stmt.getTarget().getPath());
    
    for (final Expression index : reversedPath) {
      final Term indexTerm = exprUtil.translateTerm(index);
      value = new ArrayUpdate(array, indexTerm, value);
      array = new ArrayAccess(array, indexTerm);
    }
    
    final Substitution subst = new Substitution();
    subst.setTermBinding(target.getName(), value);
    final DocumentedFormula weakestPrecondition = 
        new DocumentedFormula(subst.perform(
                               postcondition.getFormula()
                              ), 
                              postcondition.getName(), 
                              postcondition.getSource(), 
                              postcondition.getComment() + 
                                "(after substitution)");

    final VerificationCondition condition =  
        new VerificationCondition(
                ImmutableList.copyOf(preconditions), 
                weakestPrecondition, context); 
      consumer.processVerificationCondition(condition);      
  }

  /**
   * Process a skip statement.
   * @param context Context information for created conditions.
   * @param preconditions Preconditions.
   * @param stmt Statement.
   * @param postconditon Postcondition.
   * @param consumer Object used process generated conditions.
   */
  private void processSkip(
      final VCContext context,
      final Iterable<DocumentedFormula> preconditions,
      final Skip stmt,
      final DocumentedFormula postcondition,
      final VCConsumer consumer) {
      final VerificationCondition condition =  
        new VerificationCondition(ImmutableList.copyOf(preconditions), 
                                   postcondition,
                                   context); 
      consumer.processVerificationCondition(condition);    
  }
//
//  /**
//   * Create a context for a statement / postcondition pair.
//   * @param parent Parent context.
//   * @param stmt Statement.
//   * @param assertion Postcondition of stmt.
//   * @return Context object.
//   */
//  private VCContext getContext(final VCContext parent, 
//                               final SimpleStatement stmt,
//                               final Expression assertion) {
//    final String name = getAssertionName(assertion);
//    final String comment = getAssertionComment(assertion);
//
//    final List<EObject> source = new ArrayList<EObject>();
//    if (assertion.eContainer() != null) {
//      source.add(assertion.eContainer());
//    } else {
//      source.add(assertion);
//    }
//    
//    return new VCContext(name, source, comment, parent);
//  }
  
}
