package net.afyre.haha.demo.vcgen;

import java.util.ArrayList;
import java.util.List;

import net.afyre.haha.demo.haha.Add;
import net.afyre.haha.demo.haha.ArrayAccess;
import net.afyre.haha.demo.haha.ArrayType;
import net.afyre.haha.demo.haha.BinaryOp;
import net.afyre.haha.demo.haha.BoolLiteral;
import net.afyre.haha.demo.haha.Call;
import net.afyre.haha.demo.haha.Conjunction;
import net.afyre.haha.demo.haha.Disjunction;
import net.afyre.haha.demo.haha.Div;
import net.afyre.haha.demo.haha.Eq;
import net.afyre.haha.demo.haha.Equivalence;
import net.afyre.haha.demo.haha.Exists;
import net.afyre.haha.demo.haha.Forall;
import net.afyre.haha.demo.haha.Ge;
import net.afyre.haha.demo.haha.Gt;
import net.afyre.haha.demo.haha.Implication;
import net.afyre.haha.demo.haha.IntLiteral;
import net.afyre.haha.demo.haha.Le;
import net.afyre.haha.demo.haha.Lt;
import net.afyre.haha.demo.haha.Mod;
import net.afyre.haha.demo.haha.Mult;
import net.afyre.haha.demo.haha.Negation;
import net.afyre.haha.demo.haha.Neq;
import net.afyre.haha.demo.haha.Power;
import net.afyre.haha.demo.haha.Predicate;
import net.afyre.haha.demo.haha.QuantifiedVar;
import net.afyre.haha.demo.haha.ResultVar;
import net.afyre.haha.demo.haha.SimpleType;
import net.afyre.haha.demo.haha.Sub;
import net.afyre.haha.demo.haha.TypeUtils;
import net.afyre.haha.demo.haha.Uminus;
import net.afyre.haha.demo.haha.UnaryOp;
import net.afyre.haha.demo.haha.Var;
import net.afyre.haha.demo.haha.util.HahaSwitch;
import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionFactory;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.impl.SimpleExpressionFactory;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

//TODO: Translations of booleans is incorrect...
//TODO: Operation on Ints are not handled correctly...
/**
 * Switch used to translate AST expressions to SMT terms, types and formulae.
 * Note that it might be necessary to cast the returned object to 
 * a proper expression subclass.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class ExpressionSwitch extends HahaSwitch<Expression> {
  /** Factory used to construct expressions. */
  private final ExpressionFactory factory = new SimpleExpressionFactory();
  
  /** Subswitch used to handle unary operators. */
  private final UnaryOpSwitch unarySwitch = new UnaryOpSwitch();

  /** Subswitch used to handle binary operators. */
  private final BinaryOpSwitch binarySwitch = new BinaryOpSwitch();
  
  /** @{inheritDoc} */
  @Override
  public Expression caseUnaryOp(final UnaryOp expr) {
    unarySwitch.setArg(doSwitch(expr.getArg()));
    return unarySwitch.doSwitch(expr);
  }
  
  /** @{inheritDoc} */
  @Override
  public Expression caseBinaryOp(final BinaryOp expr) {
    final Expression left = doSwitch(expr.getLeft());
    final Expression right = doSwitch(expr.getRight());
    binarySwitch.setLeft(left); 
    binarySwitch.setRight(right);
    return binarySwitch.doSwitch(expr);
  }
  
  /** @{inheritDoc} */
  @Override
  public Expression caseSimpleType(final SimpleType type) {
    switch (type.getType()) {
      case BOOLEAN: return factory.createBitvector(1);
      case INT: return factory.createBitvector(TypeUtils.INT_SIZE);
      case Z: return factory.createInt();
      default: throw new RuntimeException("Unknown primitive type: " + type);
    }
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseArrayType(final ArrayType arrayType) {
    final ValueType elementType = 
        (ValueType)doSwitch(arrayType.getElementType());
    return factory.createArray(factory.createInt(), elementType);
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseIntLiteral(final IntLiteral literal) {
    return factory.createIntConst(literal.getValue());
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseBoolLiteral(final BoolLiteral literal) {
    switch (literal.getValue()) {
      case FALSE: return factory.createFalse();
      case TRUE: return factory.createTrue();
      default:
        assert false;
        return null;
    }
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseForall(final Forall forall) {
    final List<VariableBinding> bindings = new ArrayList<VariableBinding>();
    for (final QuantifiedVar variable : forall.getVars()) {
      final ValueType varType = (ValueType)doSwitch(variable.getType());
      final VariableBinding binding = 
          new VariableBinding(variable.getName(), varType);
      bindings.add(binding);
    }
    final Formula arg = (Formula)doSwitch(forall.getArg());
    final Expression[] patterns = 
        new Expression[forall.getPatterns().size()];
    for (int i = 0; i < patterns.length; i++) {
      patterns[i] = doSwitch(forall.getPatterns().get(i));
    }
    return factory.createForall(bindings, arg, patterns);
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseExists(final Exists exists) {
    final List<VariableBinding> bindings = new ArrayList<VariableBinding>();
    for (final QuantifiedVar variable : exists.getVars()) {
      final ValueType varType = (ValueType)doSwitch(variable.getType());
      final VariableBinding binding = 
          new VariableBinding(variable.getName(), varType);
      bindings.add(binding);
    }
    final Formula arg = (Formula)doSwitch(exists.getArg());
    final Expression[] patterns = 
        new Expression[exists.getPatterns().size()];
    for (int i = 0; i < patterns.length; i++) {
      patterns[i] = doSwitch(exists.getPatterns().get(i));
    }
    return factory.createExists(bindings, arg, patterns);
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseVar(final Var var) {
    return factory.createVariable(var.getName());
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseResultVar(final ResultVar var) {
    return factory.createVariable(var.getName());
  }
  
  /** @{inheritDoc} */
  @Override
  public Expression caseCall(final Call call) {
    final String pred = call.getFunction().getName();
    final List<Term> args = new ArrayList<Term>();
    for (final net.afyre.haha.demo.haha.Expression arg : 
           call.getArgs().getArguments()) {
      args.add((Term)doSwitch(arg));
    }
    if (call.getFunction() instanceof Predicate) {
      return factory.createPredicateCall(pred + "?", args.toArray(new Term[]{}));
    } else {
      return factory.createCall("function@" + pred, args.toArray(new Term[]{}));
    }
  }

  /** @{inheritDoc} */
  @Override
  public Expression caseArrayAccess(final ArrayAccess expr) {
    final Term array = (Term)doSwitch(expr.getArray());
    Term result = array;
    for (final net.afyre.haha.demo.haha.Expression index : expr.getPath()) {
      final Term translatedIndex = (Term)doSwitch(index);
      result = factory.createArrayAccess(result, translatedIndex);
    }
    return result;
  }
  
  /**
   * Switch used to handle unary operators.
   * 
   * Before the call to 'doSwitch', the argument is translated 
   * and passed in the 'arg' field.
   */
  private class UnaryOpSwitch extends HahaSwitch<Expression> {
    /** Translated argument. */
    private Expression arg;
    
    /**
     * Set the translated argument (used before doSwitch()). 
     * @param arg Processed argument.
     */
    public void setArg(final Expression arg) {
      this.arg = arg;
    }
    
    /** @{inheritDoc} */
    @Override
    public Expression caseNegation(final Negation neg) {
      return factory.createNegation((Formula)arg);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseUminus(final Uminus expr) {
      return factory.createUminus((Term)arg);
    }
  }
  
  /**
   * Switch used to handle binary operators.
   * 
   * Before the call to 'doSwitch', both arguments are translated 
   * and passed in the 'left' and 'right' fields.
   */  
  private class BinaryOpSwitch extends HahaSwitch<Expression> {
    /** Translated left argument. */
    private Expression left;
    
    /** Translated right argument. */
    private Expression right;
    
    /**
     * Set the left translated argument (used before doSwitch()). 
     * @param arg Processed argument.
     */
    public void setLeft(final Expression left) {
      this.left = left;
    }
    
    /**
     * Set the right translated argument (used before doSwitch()). 
     * @param arg Processed argument.
     */    
    public void setRight(final Expression right) {
      this.right = right;
    }
    
    /** @{inheritDoc} */
    @Override
    public Expression caseEquivalence(final Equivalence equiv) {
      return factory.createEquivalence((Formula)left, (Formula)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseImplication(final Implication impl) {
      return factory.createImplication((Formula)left, (Formula)right);
    }
    
    /** @{inheritDoc} */
    @Override
    public Expression caseDisjunction(final Disjunction expr) {
      return factory.createDisjunction((Formula)left, (Formula)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseConjunction(final Conjunction expr) {
      return factory.createConjunction((Formula)left, (Formula)right);
    }
    
    /** @{inheritDoc} */
    @Override
    public Expression caseEq(final Eq eq) {
      return factory.createEq((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseNeq(final Neq neq) {
      return factory.createNeq((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseLt(final Lt lt) {
      return factory.createLt((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseLe(final Le le) {
      return factory.createLe((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseGt(final Gt gt) {
      return factory.createGt((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseGe(final Ge ge) {
      return factory.createGe((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseAdd(final Add add) {
      return factory.createAddition((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseSub(final Sub sub) {
      return factory.createSubtraction((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseMult(final Mult mult) {
      return factory.createMultiplication((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseDiv(final Div div) {
      return factory.createDivision((Term)left, (Term)right);
    }

    /** @{inheritDoc} */
    @Override
    public Expression caseMod(final Mod mod) {
      return factory.createRemainder((Term)left, (Term)right);
    }
    
    /** @{inheritDoc} */
    @Override
    public Expression casePower(final Power power) {
      return factory.createPower((Term)left, (Term)right);
    }
  }
}
