package net.afyre.haha.demo.vcgen;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import org.eclipse.emf.ecore.EObject;

import net.afyre.haha.demo.haha.Argument;
import net.afyre.haha.demo.haha.Call;
import net.afyre.haha.demo.haha.Callable;
import net.afyre.haha.demo.haha.Expression;
import net.afyre.haha.demo.haha.Function;
import net.afyre.haha.demo.haha.Postcondition;
import net.afyre.haha.demo.haha.Precondition;
import net.afyre.haha.demo.solver.context.AxiomDeclaration;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.expr.ExpressionFactory;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.impl.SimpleExpressionFactory;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.util.Substitution;
import net.afyre.haha.demo.util.Pair;
import net.afyre.haha.demo.util.ZipIterable;

/**
 * Utilities used by various VCGen components.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public final class VCGenUtil {
  /** Singleton instance. */
  private static VCGenUtil instance;

  /** Object used to convert AST expressions to terms. */
  private final ExpressionUtil exprUtil = ExpressionUtil.getInstance();
  
  /** Object used to create expressions. */
  private final ExpressionFactory exprFactory = 
      new SimpleExpressionFactory();
  
  /**
   * Create an instance of this class.
   * @return Instance.
   */
  public synchronized static VCGenUtil getInstance() {
    if (instance == null) {
      instance = new VCGenUtil();
    }
    return instance;
  }
  
  /** Constructor (hidden). */
  private VCGenUtil() { }
  
  /**
   * Create a term representing a function call where
   * actual arguments are variables with names identical
   * to the formal arguments.
   * 
   * @param function Function definition.
   * @return Term.
   */
  public Term getFunctionCall(final Function function) {
    final List<Term> actualArgs = new ArrayList<Term>(
        function.getArguments().getArguments().size());
    for (final Argument arg : function.getArguments().getArguments()) {
      actualArgs.add(exprFactory.createVariable(arg.getName()));
    }
    return exprFactory.createCall("function@" + function.getName(), 
                                  actualArgs.toArray(new Term[]{}));
  }
  
  /**
   * Extract topmost function calls from an AST node.
   * Nested calls should *not* be returned.
   * 
   * @param expr Expression to process.
   * @return List of calls.
   */
  public List<Call> getTopmostCalls(final EObject expr) {
    final List<Call> result = new ArrayList<Call>();
    final Stack<EObject> stack = new Stack<EObject>();
    
    stack.push(expr);
    loop: while (!stack.isEmpty()) {
      final EObject current = stack.pop();
      if (current instanceof Call) {
        final Call call = (Call)current;
        // Do not return predicate calls
        if (call.getFunction() instanceof Function) {
          result.add(call);
          // Skip child nodes.
          continue loop;
        }
      }
      for (final EObject child : current.eContents()) {
        stack.push(child);
      }
    }
    return result;
  }
  
  // TODO :COMMENT, SPLIT !!!!!!!!!!!!!!!!!!!!
  public VCContext processFunctionCalls(final VCContext context,
                                        final List<DocumentedFormula> preconditions,
                                        final EObject expr, 
                                        final VCConsumer consumer) {
    final List<Call> calls = getTopmostCalls(expr);
    if (calls.isEmpty()) {
      // No calls, just return the original context.
      return context;
    }
    for (final Call call : calls) {
      final Function function = (Function)call.getFunction();
      
      // Process nested calls
      final VCContext callContext = 
          processFunctionCalls(context, preconditions, call.getArgs(), consumer);
      
      // Prove preconditions
      final Substitution subst = getCallSubst(call);
      for (final Precondition pre : function.getPreconditions()) {
        final Formula preExpr = 
            subst.perform(exprUtil.translateFormula(pre.getFormula()));
        final List<EObject> sources = new ArrayList<EObject>();
        sources.add(pre);
        sources.add(call);
        final DocumentedFormula df = 
            new DocumentedFormula(preExpr, 
                ExprToDocumentedFormula.getAssertionName(pre.getFormula()), 
                sources, "Precondtion of " + function.getName());
        final VerificationCondition vc = new VerificationCondition(
            preconditions, df, callContext);
        consumer.processVerificationCondition(vc);
      }
    }
    

    // Aggregate postconditions for all calls
    final List<Declaration> postconditions = new ArrayList<Declaration>();
    for (final Call call : calls) {
      final Function function = (Function)call.getFunction();
      final Substitution argSubst = getCallSubst(call);
      for (final Postcondition post : function.getPostconditions()) {
        final Substitution resultSubst = new Substitution();
        resultSubst.setTermBinding(function.getName(), 
                             getFunctionCall(function));
        Formula postExpr = exprUtil.translateFormula(post.getFormula()); 
        postExpr = resultSubst.perform(postExpr);
        postExpr = argSubst.perform(postExpr);
        final String name = 
            ExprToDocumentedFormula.getAssertionName(post.getFormula());
        final AxiomDeclaration decl = new AxiomDeclaration(name, postExpr);
        postconditions.add(decl);
      }
    }
    
    // Create new context
    final VCContext afterCalls = 
        new VCContext(context.getName() + "_after_calls", 
            context.getSource(), "After function calls", context,
            postconditions);
    
    return afterCalls;
  }

  // TODO :COMMENT!!!!!!!!!!!!!!!!!!!!!!
  public Substitution getCallSubst(final Call call) {
    final Callable function = call.getFunction();
    final Substitution subst = new Substitution();
    final List<Argument> args = function.getArguments().getArguments();
    final List<Expression> actualArgs = call.getArgs().getArguments();
    for (final Pair<Argument, Expression> pair : 
             ZipIterable.getInstance(args, actualArgs)) {
      final String arg = pair.getFirst().getName();
      final Term actualArg = exprUtil.translateTerm(pair.getSecond());
      subst.setTermBinding(arg, actualArg);
    }
    return subst;
  }
  
}
